#include "uart.h"

speed_t
uart_get_speed(int requested_speed)
{
  speed_t speed;

  switch(requested_speed)
  {
    case 0:
      speed = B0;
      break;
    case 50:
      speed = B50;
      break;
    case 75:
      speed = B75;
      break;
    case 110:
      speed = B110;
      break;
    case 134:
      speed = B134;
      break;
    case 150:
      speed = B150;
      break;
    case 200:
      speed = B200;
      break;
    case 300:
      speed = B300;
      break;
    case 600:
      speed = B600;
      break;
    case 1200:
      speed = B1200;
      break;
    case 1800:
      speed = B1800;
      break;
    case 2400:
      speed = B2400;
      break;
    case 4800:
      speed = B4800;
      break;
    case 9600:
      speed = B9600;
      break;
    case 19200:
      speed = B19200;
      break;
    case 38400:
      speed = B38400;
      break;
    case 57600:
      speed = B57600;
      break;
    case 115200:
      speed = B115200;
      break;
    case 230400:
      speed = B230400;
      break;
    case 460800:
      speed = B460800;
      break;
    case 500000:
      speed = B500000;
      break;
    case 576000:
      speed = B576000;
      break;
    case 921600:
      speed = B921600;
      break;
    case 1000000:
      speed = B1000000;
      break;
    case 1152000:
      speed = B1152000;
      break;
    case 1500000:
      speed = B1500000;
      break;
    case 2000000:
      speed = B2000000;
      break;
    default:
      speed = -1;
  }

  return speed;
}

int
uart_setup(char *filename, speed_t baud, bool parity, bool twostop)
{
  int UART;
  struct termios tty_orig;

  if(access(filename, F_OK) == -1)
  {
    err_output("unable to access path %s\n", filename);
    return -2;
  }

  UART = open(filename, O_RDWR | O_NOCTTY);
  if(UART == -1)
  {
    err_output("error opening terminal %s\n", filename);
    close(UART);
    return UART;
  }

  if(tcgetattr(UART, &tty_orig) == -1)
  {
    err_output("unable to get tty attributes for %s\n", filename);
    close(UART);
    return -1;
  }

  if(cfsetispeed(&tty_orig, baud) == -1)
  {
    errno_output("unable to set input speed\n");
    close(UART);
    return -1;
  }
  if(cfsetospeed(&tty_orig, baud) == -1)
  {
    errno_output("unable to set output speed\n");
    close(UART);
    return -1;
  }

  cfmakeraw(&tty_orig);

  tty_orig.c_cflag |= CREAD | CLOCAL | CS8;
  tty_orig.c_cc[VMIN] = 0;
  tty_orig.c_cc[VTIME] = 0;

  if(parity)
  {
    tty_orig.c_cflag |= PARENB;
    tty_orig.c_iflag |= PARMRK | INPCK;
  }
  else
  {
    tty_orig.c_iflag |= IGNPAR;
  }

  if(twostop)
  {
    tty_orig.c_cflag |= CSTOPB;
  }
  else
  {
    tty_orig.c_cflag &= ~(CSTOPB);
  }

  if(tcsetattr(UART, TCSANOW, &tty_orig) == -1)
  {
    err_output("error setting terminal attributes\n");
    close(UART);
    return -1;
  }

  tcflush(UART, TCIFLUSH);
  tcdrain(UART);

  return UART;
}


static void
uart_poll_init(int fd_uart, struct pollfd pfd[])
{
  bool istty;

  istty = isatty(fileno(stdin));
  if(!istty)
  {
    pfd[0].fd = -1;
    pfd[0].events = 0;
  }
  else
  {
    pfd[0].fd = fileno(stdin);
    pfd[0].events = POLLIN;
  }

  pfd[1].fd = fd_uart;
  pfd[1].events = POLLIN;
}

static ssize_t
uart_write(int fd_uart, bool verbose)
{
  ssize_t rbytes, wbytes;
  char buf[8];

  rbytes = read(fileno(stdin), buf, 7);
  if(rbytes == -1)
  {
    errno_output("error reading from stdin\n");
    return rbytes;
  }
  buf[rbytes] = '\0';

  if(verbose)
  {
    debug_output("stdin read[%d]: %s", rbytes, buf);
    for(int i=0;i<rbytes;i++)
      debug_output("%d %d\n", i, buf[i]);
  }

  wbytes = write(fd_uart, buf, rbytes);
  if(wbytes == -1)
  {
    errno_output("error writing to uart\n");
    return -2;
  }

  if(verbose)
    debug_output("uart  write[%d]: %s", wbytes, buf);

  if(wbytes != rbytes)
  {
    warn_output("wrote only %d of %d bytes\n", wbytes, rbytes);
  }

  return rbytes;
}

static ssize_t
uart_read(int fd_uart, char buffer[], size_t buf_len, bool verbose)
{
  ssize_t rbytes;

  rbytes = read(fd_uart, buffer, buf_len);
  if(rbytes ==  -1)
  {
    errno_output("error reading from UART\n");
    return rbytes;
  }

  return rbytes;
}

int
uart_poll(int fd_uart, bool verbose)
{
  int ret;
  struct pollfd pfd[2];
  uart_poll_init(fd_uart, pfd);
  char buffer[8];
  ssize_t bytes_read;

  while(true)
  {
    ret = poll(pfd, 2, -1);
    if(ret == 0)
    {
      err_output("poll timed out\n");
      return 1;
    }
    else if(ret < 0)
    {
      errno_output("poll\n");
      return 2;
    }

    if(verbose)
    {
      debug_output("received input\n");
    }

    if(pfd[0].revents & POLLIN)
    {
      uart_write(fd_uart, verbose);
    }

    if(pfd[1].revents & POLLIN)
    {
      bytes_read = uart_read(fd_uart, buffer, 8, verbose);
      if(bytes_read > 0)
      {
        info_output("uart  read[%d]: %s", bytes_read, buffer);
      }
    }
  }

  return 0;
}
